namespace NXyn.Aux
    open System

    module ListMisc = 
        let rec rev_append (xs:list<'a>) (ys:list<'a>) : list<'a> =
                           match xs
                            with | [] -> ys
                                 | x::xs -> rev_append xs (x::ys)

        let member_of (x:'a) (xs:list<'a>) : Boolean =
                      List.exists ((=)x) xs

        let eliminateNeighboringDups (xs:list<'a>) : list<'a> =
                                     let rec loop bag = function [] -> List.rev(bag)
                                                               | x1::xs -> loop (match xs with x2::_ when x1 = x2 -> bag | _ -> x1::bag) xs
                                     loop [] xs

        let insert_into_sorted (p:'a -> 'a -> Int32) (el:'a) (xs:list<'a>) : list<'a> =
                               let rec loop xxs =
                                            match xxs
                                             with | [] -> [el]
                                                  | x::xs -> if p el x < 0 then el::xxs else x::(loop xs)
                               loop xs

        let select_first (p:'a -> Boolean) (xs:list<'a>) : option<list<'a> * 'a * list<'a>> =
                         let rec loop pre' xxs =
                                      match xxs
                                       with | [] -> None
                                            | x::xs -> if p x then Some(pre',x,xs) else loop (x::pre') xs
                         loop [] xs

        let select_first' (f:list<'a> -> 'a -> list<'a> -> option<'b>) (xs:list<'a>) : option<'b> =
                          let rec loop f pre' = function [] -> None
                                                       | x::xs -> match f pre' x xs
                                                                   with | None -> loop f (x::pre') xs
                                                                        | Some _ as res -> res
                          loop f [] xs

                         
        let inline cons (x:'a) (xs:list<'a>) = x::xs
        
        let inline revCons (xs:list<'a>) (x:'a) = x::xs


        (* split one list to two lists, both having roughly the same length *)
        let zigzag (xs:'a list) : ('a list * 'a list) =
                   let rec loop ls rs = function []    -> (ls,rs)
                                               | x::xs -> loop (x::rs) ls xs
                   loop [] [] xs
