namespace NXyn.EmitFSharp

  open System
  open System.Reflection
  open Microsoft.FSharp.Reflection
  open NXyn.EmitFSharp.Common
  
  module MapUnionToUnion =
    let OutputCode (w:System.CodeDom.Compiler.IndentedTextWriter) (tFrom:System.Type) (tTo:System.Type) (name:string) =
      if tFrom.ContainsGenericParameters || tTo.ContainsGenericParameters then failwith "EmitFSharpShallowUnionMapper.OutputCode: open types unsupported"

      w.WriteLine("let {0} (arg : {1}) : Option<{2}> =", name, FSharpTypeName tFrom, FSharpTypeName tTo)
      w.Indent <- w.Indent + 1
      w.WriteLine("match arg with")
      w.Indent <- w.Indent + 1

      let (tFromShortName,tToShortName) = (ClippedFSharpTypeName tFrom.Name, ClippedFSharpTypeName tTo.Name)
      let (tFromCases, tToCases) = (FSharpType.GetUnionCases tFrom, FSharpType.GetUnionCases tTo)

      match Array.partition (fun (uc:UnionCaseInfo) -> Array.exists (UnionCaseInfo_equal uc) tToCases) tFromCases with
        | ([||], _) -> w.WriteLine("| _ -> None")
        | (ok,no) ->
            for (uc:UnionCaseInfo) in ok do
              match uc.GetFields () with
                | [||] -> w.WriteLine("| {0}.{2} -> Some({1}.{2})", tFromShortName, tToShortName, uc.Name)
                | fs ->
                    WriteList     w ("| " + tFromShortName + "." + uc.Name + "(")  "," ") -> " fs.Length (fun i -> "a" + string i)
                    WriteListLine w ("Some(" + tToShortName + "." + uc.Name + "(") "," "))"    fs.Length (fun i -> "a" + string i)

            for (uc:UnionCaseInfo) in no do
              w.WriteLine("| {0} -> None", tFromShortName + "." + uc.Name + if uc.GetFields () = [||] then "" else " _")
      w.Indent <- w.Indent - 2
      w.WriteLine(";;\n")
