

module GlobalScheduling


open System
open System.Collections.Generic

open Ast
open X64instr

open InstrIO
open MoreInstrIO

open AstIO
open MoreAstIO


open NXyn
open NXyn.Aux
open NXyn.Aux.Global

open NXyn.Aux.IntMisc
open NXyn.Aux.OtherMisc
open NXyn.Aux.SetMisc


open Basics

open LocalScheduling

open OptSimple


let get_instrs_info (schedinfo:ListScheduling.Parameters) (b:IBlock) (in0:Set<Operand>) (out0:Set<Operand>) (instrs:Instr []) =

//            let (tLast,_) = instrs.[instrs.Length - 1]

            let dead = Array.create instrs.Length Set.empty

            let iSrcs = Array.create instrs.Length Set.empty
            let iDsts = Array.create instrs.Length Set.empty

            // XXX take vec regs only?
            instrs |> Array.iteri (fun ix instr -> let (s,ds,d) = (S_of_Instr konstSizeDict b.ReturnsVoid instr, DS_of_Instr instr, D_of_Instr instr)
                                                   iSrcs.[ix] <- Set.union s ds
                                                   iDsts.[ix] <- Set.union ds d)

            let allDsts = Array.fold Set.union Set.empty iDsts
//            let allSrcs = Array.fold Set.union Set.empty iDsts
            let all_d = ref allDsts //(Set.union allSrcs allDsts)
//            if allDsts <> !all_d then failwith "OJE"
            
//            let srcs_dsts = instrs |> Array.map (fun(_,instr) -> let (s,ds,d) = (Get_S_of_Instr b instr, Get_DS_of_Instr b instr, Get_D_of_Instr b instr)
//                                                                 (Set.union s ds, Set.union ds d))

            let iLatency = instrs |> Array.map schedinfo.GetLatency

//            let t_read = new Dictionary<Operand, Int32>()
//            for x in out0 do t_read.[x] <- tLast + 1

//            let t_first_output_read = Array.create instrs.Length None // (tLast+1)

            Array.foldBack (fun instr (ix,maybedead) ->
                              let dsts = D_of_Instr instr
                              let dstsDS = Set.union dsts (DS_of_Instr instr)
//                              let dsts   = iDsts.[ix] // Set.union (Get_D_of_Instr b instr) (Get_DS_of_Instr b instr)
//                              let dstsDS = dsts // Set.union (Get_D_of_Instr b instr) (Get_DS_of_Instr b instr)

                              let srcs = iSrcs.[ix] // Set.union (Get_S_of_Instr b instr) (Get_DS_of_Instr b instr)

                             
//                              if not(dstsDS.IsEmpty) then
//                                let sel = dstsDS |> Set.toList |> List.choose (fun(operand) -> if t_read.ContainsKey(operand) then Some(t_read.[operand]) else None)
//                                if sel <> [] then t_first_output_read.[ix] <- Some(List.min(sel))
                            
//                                let not_found = dsts |> Set.toList |> List.filter (fun(operand) -> not(t_read.ContainsKey(operand)))
//                                if not_found = [] then 
//                                else not_found |> List.iter (fun(operand) -> eprintfn "not(t_read.ContainsKey(operand)) : %A, ix = %d, instr = %A" operand ix instr)
                            
//                              for src in srcs do t_read.[src] <- t

                              let maybedead = Set.union      maybedead dsts  (* XXX dstsDS? *)
                              let maybedead = Set.difference maybedead out0
                              let deadhere  = Set.intersect  maybedead srcs
                              dead.[ix] <- deadhere
                              let maybedead = Set.difference maybedead deadhere
                              (ix-1,maybedead)
                              ) instrs (instrs.Length - 1, Set.difference !all_d out0)
            |> fun(ix0,dead0) ->
                

//                let t_written = new Dictionary<Operand, Int32>()
//                for x in in0 do t_written.[x] <- (-1)

//                let t_last_input_written = Array.create instrs.Length None            // TBD should be t_last_input_ready
            
//                for ix = 0 to instrs.Length - 1 do
//                    let (t,instr) = instrs.[ix]
//                    let srcs = iSrcs.[ix]
//                    let dsts = iDsts.[ix]
//
//                    if not(srcs.IsEmpty) then t_last_input_written.[ix] <- Some(List.max (srcs |> Set.toList |> List.map (fun(operand) -> t_written.[operand])))
//
//                    for dst in dsts do t_written.[dst] <- t + iLatency.[ix]

                (iSrcs,iDsts,iLatency,dead,dead0)
//                 t_read,t_written,
//                 t_last_input_written,t_first_output_read)




// simplify! move down maximally, NO LOOP
// which args are no longer needed?
let loopMD schedinfo (b:IBlock) (* (in0:Set<Operand>) (out0:Set<Operand>) *) (instrs:Instr []) =
            instrs
            |> Array.toList
            |> moveDown_Instrs_like b (function XsXd(_) | YsYd(_) -> true
                                              | M16sXd(_) | M32sXd(_) | M64sXd(_) | M128sXd(_) | M32sYd(_) | M64sYd(_) | M128sYd(_) | M256sYd(_) -> true
                                              | M8sXds(_) | M16sXds(_) | M32sXds(_) | M64sXds(_) | M128sXds(_) -> true
                                              | R64d(_) -> true
                                              | R32sR32ds(_) | R64sR64ds(_) | R64sR64d(_) -> true
                                              | _ -> false)
            |> Array.ofList


(***)

// what about fixedX?

// (stayInHalf:Boolean)

let regreallocV schedinfo (b:IBlock) (in0:Set<Operand>) (out0:Set<Operand>) (instrs:Instr []) =

            (* get all vector registers that are actually written in this block *)
(*            let all_d = ref Set.empty

            for instr in instrs do
                all_d := Set.union !all_d (Get_D_of_Instr  instr)
                all_d := Set.union !all_d (Get_DS_of_Instr instr)
                *)
//                all_d := Set.union !all_d (Get_S_of_Instr konstSizeDict b.ReturnsVoid instr) // NEW

//            for instr in b.Instrs do
// /               all_d := Set.union !all_d (Get_D_of_Instr  instr)
//                all_d := Set.union !all_d (Get_DS_of_Instr instr)



            (* realloc X registers [TBD should be same for Y registers, too] *)
            let (iSrcs,iDsts,iLatency,dead,dead0) = get_instrs_info schedinfo b in0 out0 instrs
//                 t_read,t_written,
//                 t_last_input_written,t_first_output_read) = 
            (* MOSTLY DEL, we don't need most of that info ... *)


            let substVregs = new Dictionary<Int32,Int32>() (* index -> index *)

            for x in in0 do
                match x with Vector(x) -> substVregs.[x] <- x
                           | _ -> ()
            for x in out0 do
                match x with Vector(x) -> substVregs.[x] <- x
                           | _ -> ()

            // XXX
            let free = dead0 (* Set.difference !all_d out0 *) |> Set.toList |> List.choose (function Vector(x) -> Some(x) | _ -> None)
//            let free = Set.difference !all_d (Set.union in0 out0) |> Set.toList |> List.choose (function Vector(x) -> Some(x, -1) | _ -> None)

            // (in,out) could include fixed registers (HACK) to keep them from being reallocated, etc.


            let stayInHalf = reassign_xmm() <> ReassignXMM.No // TargetIsIntel() // true //Basics.target = Intel // true // false

//            let nfreeMin = ref 0

            let newinstrs = new ResizeArray<Instr>()
            let free' = Array.fold (fun (ix,free0) instr0 ->
                                      let instr = Map_Xds_and_Xs_in_Instr (fun(x:X) -> X128(substVregs.[x.Int])) instr0
                                      let instr = Map_Yds_and_Ys_in_Instr (fun(x:Y) -> Y256(substVregs.[x.Int])) instr
//                                      let instr = Map_Xds_in_Instr (fun(x:X) -> X128(substVregs.[x.Int])) instr
//                                      let instr = Map_Yds_in_Instr (fun(x:Y) -> Y256(substVregs.[x.Int])) instr

//                                      nfreeMin := !nfreeMin + (List.length free0)
                                      let free = ref free0

                                      let dead = dead.[ix] |> fun s -> Set.difference s out0 |> Set.toList |> List.choose (function Vector(v) -> Some(v) | _ -> None)
                                      
                                      for r in dead do //.[ix] do
                                            if not(substVregs.ContainsKey(r)) then failwithf "%s: realloc err222" !Global.InFileName
                                            free := !free @ [substVregs.[r]]
                                            substVregs.Remove(r) |> ignore


                                      let dsts = D_of_Instr instr |> fun s -> Set.difference s out0 |> Set.toList |> List.choose (function Vector(v) when not(substVregs.ContainsKey(v)) -> Some(v) | _ -> None)

//                                          match r with Vector(r) -> free := (r,t)::!free
//                                          match r with
//                                              | Vector(r) -> if not(Set.contains (Vector r) out0) then
//                                              | _ -> ()

                                      for r in dsts do
                                          let f = match !free
                                                   with | [] -> for i = ix - 30 to ix+30 do eprintfn "[#%d]: %A, dead = %A" i instrs.[i] dead.[i]
                                                                failwithf "regrealloc(X): no free registers %s: [%d] %A" !Global.InFileName ix instr0
                                                        | f::_ when not(stayInHalf) -> f
                                                        | f::_ -> defaultArg (List.tryFind (fun(f) -> f / 8 = r / 8) !free) f
                                          
                                          substVregs.[r] <- f
                                          free := !free |> List.filter ((<>)f)

                                      let instr = Map_Xd_in_Instr (fun(x:X) -> X128(substVregs.[x.Int])) instr
                                      let instr = Map_Yd_in_Instr (fun(x:Y) -> Y256(substVregs.[x.Int])) instr
                                      newinstrs.Add(instr)
//                                      eprintfn "[%d] free0 %A => free %A" ix free0 !free
                                      (ix+1,!free))
                                   (0,free) instrs

//            eprintfn "nfreeMin = %d" !nfreeMin
            newinstrs.ToArray()



let doit (schedinfo:ListScheduling.Parameters) (inoutDict:Dictionary<String,Set<Operand> * Set<Operand>>) (tDict:Dictionary<string, IBlock>) (comp:String []) =
 



    for lbl in comp do
      let b = tDict.[lbl]
      if b.Instrs.Length > 0 then 
        
//        eprintfn "  lbl %s" lbl

        let in_out = inoutDict.[lbl]
//        realloc_stack_in_block in_out b


//        let (fwdeps,bwdepscnt,cplens) = ListScheduling.calc_dependencies schedinfo b.Instrs

//        for ix = 0 to b.Instrs.Length - 1 do
//            eprintfn "[%d] %A: cplen %d, fwdeps %A" ix b.Instrs.[ix] cplens.[ix] fwdeps.[ix]

        let dumphere = false
        if dumphere then eprintfn "before llsched ---"

        let t0 = DateTime.Now


        let rec loopSR (prevOpt:option<Int32 * (Instr [])>) (instrs:Instr []) =
            let (in0,out0) = inoutDict.[lbl]
// (* if realloc is OK *)            let instrs = swap_copy_operands (in0,out0) instrs


// (* if realloc is OK *)           let instrs = OptForwardStoreToLoad.do_opt b in0 out0 instrs

            b.Instrs <- instrs
            elim_useless_vec_copies_block (in0,out0) b
            let instrs = b.Instrs 
            let res = ListScheduling.llsched schedinfo konstSizeDict instrs

            let (t,_) = res.[res.Length - 1]
            
            eprintfn "t = %d" t

            match prevOpt with
                | Some(t0,instrs0) when t0 <= t -> instrs0
                | _ ->
                    let (in0,out0) = inoutDict.[lbl]

                    let instrsR = res |> Array.map (fun(t,ix) -> instrs.[ix])

                    b.Instrs <- loopMD schedinfo b instrsR
                    
// (* if realloc is OK *)                    elim_useless_vec_copies_block (in0,out0) b
                    
//                    let instrs = b.Instrs |> Array.map (fun(i) -> (1,i)) // PSEUDO
                 
                    let reallocd_instrs = regreallocV schedinfo b in0 out0 b.Instrs // instrs



                    loopSR (Some(t,instrsR)) reallocd_instrs
                    
        b.Instrs <- loopSR None b.Instrs

