module LocalScheduling

open System
open System.Collections.Generic

open Ast
open X64instr

open InstrIO
open MoreInstrIO

open AstIO
open MoreAstIO


open NXyn
open NXyn.Aux
open NXyn.Aux.Global

open NXyn.Aux.IntMisc
open NXyn.Aux.OtherMisc
open NXyn.Aux.SetMisc


open Basics

(* operates on list that is in-order *)
let moveDown_Instr_until stop (i0,aux0,s0,ds0,d0) mapI instrs =
    let rec loop prevOption xxs =
                 match xxs
                  with | [] -> [mapI prevOption (i0,aux0,s0,ds0,d0) None]
                       | (i1,aux1,s1,ds1,d1)::instrs -> if   stop (i1,aux1)                                                           then (mapI prevOption (i0,aux0,s0,ds0,d0) (Some(i1,aux1,s1,ds1,d1)))::(i1,aux1,s1,ds1,d1)::instrs
                                                        elif                          sets_intersect s0  ds1 || sets_intersect s0  d1 ||
                                                             sets_intersect ds0 s1 || sets_intersect ds0 ds1 || sets_intersect ds0 d1 ||
                                                             sets_intersect d0  s1 || sets_intersect d0  ds1 || sets_intersect d0  d1 then (mapI prevOption (i0,aux0,s0,ds0,d0) (Some(i1,aux1,s1,ds1,d1)))::(i1,aux1,s1,ds1,d1)::instrs
                                                                                                                                      else (i1,aux1,s1,ds1,d1)::(loop (Some (i1,aux1,s1,ds1,d1)) instrs)
    loop None instrs

let moveDown_Instr (instr,aux,s,ds,d) instrs =
                   moveDown_Instr_until (fun(instr,aux) -> Instr_is_J(instr)) (* (fun _ -> false) *) (instr,aux,s,ds,d) (fun _ x _ -> x) instrs


let moveDown_Instrs_like' (b:IBlock) (test:Instr * 'a -> Boolean) (instrs:list<Instr * 'a>) =
                          let instrs = instrs |> List.map (fun (instr,aux) -> let s  = S_of_Instr  konstSizeDict b.ReturnsVoid instr
                                                                              let ds = DS_of_Instr(instr)
                                                                              let d  = D_of_Instr(instr)
                                                                              (instr,aux,s,ds,d))
                          let instrs = List.foldBack (fun (instr,aux,s,ds,d) instrs -> if test(instr,aux) then moveDown_Instr (instr,aux,s,ds,d) instrs else (instr,aux,s,ds,d)::instrs) instrs []
                          instrs |> List.map (fun (instr,aux,s,ds,d) -> (instr,aux))


let moveDown_Instrs_like (b:IBlock) (test:Instr -> Boolean) (instrs:list<Instr>) =
                         let instrs = instrs |> List.map (fun(i) -> (i,None))
                         let instrs = moveDown_Instrs_like' b (fun(i,_) -> test(i)) instrs
                         instrs |> List.map (fun (i,_) -> i)
                         
let moveUp_Instrs_like b test instrs =
                       let instrs = instrs |> List.map (fun instr -> (instr,None))
                       let instrs = instrs |> List.map (fun (instr,aux) -> let s  = S_of_Instr  konstSizeDict b.ReturnsVoid instr
                                                                           let ds = DS_of_Instr(instr)
                                                                           let d  = D_of_Instr(instr)
                                                                           (instr,aux,s,ds,d))
                       let rec loop bag = function
                           | [] -> List.rev(bag)
                           | (instr0,aux0,s0,ds0,d0)::instrs when not(Instr_is_J(instr0)) && test (instr0,aux0,s0,ds0,d0) instrs ->
                               let bag2 = moveDown_Instr_until (* NAME *) (fun _ -> false) (instr0,aux0,s0,ds0,d0) (* instr *) (fun _ x _ -> x) bag
                               loop bag2 instrs
                           | instr::instrs -> loop (instr::bag) instrs
                       let instrs = loop [] instrs
                       instrs |> List.map (fun(instr,_,_,_,_) -> instr)

